###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = 'Donovan Parks'
__copyright__ = 'Copyright 2015'
__credits__ = ['Donovan Parks']
__license__ = 'GPL3'
__maintainer__ = 'Donovan Parks'
__email__ = 'donovan.parks@gmail.com'

import csv
import hashlib
import re


def sha256(input_file):
    """Calculate the SHA256 hash for a specific file.

    :param input_file: str, Name of file.

    :return: str, SHA256 hash.

    """

    BLOCKSIZE = 65536
    hasher = hashlib.sha1()
    with open(input_file, 'rb') as afile:
        buf = afile.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(BLOCKSIZE)

    return hasher.hexdigest()

def sha256_rb(input_file):
    """Calculate SHA256 hash for a specific file with a binary format.

    :param input_file: str, Name of file.

    :return: str, SHA256 hash.

    """

    BLOCKSIZE = 65536
    hasher = hashlib.sha1()
    with input_file as onefile:
        buf = onefile.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = onefile.read(BLOCKSIZE)

    return hasher.hexdigest()

def select_delimiter(input_file):
    """Select delimiter for a specific file (.csv,.tsv,...).

    :param input_file: str, Name of file.

    :return: str, Delimiter.

    """
    with open(input_file) as f:
        first_line = f.readline()
        # detect the separator
        sniffer = csv.Sniffer()
        dialect = sniffer.sniff(first_line)
        return dialect.delimiter

def matching_brackets(line):
    """Split a string to a list with matching brackets.
    a string a split into substrings based on matching brackets. so we can return nested brackets.

    :param line: str, line to be processed.

    :return: list, list of strings.
    """

    result, start_parens_count, end_parens_count, term = [], 0, 0, ""
    for x in re.split(r"([\[\]])", line):
        if not x.strip():
            continue
        elif x == "[":
            if start_parens_count > 0:
                term += "["
            start_parens_count += 1
        elif x == "]":
            end_parens_count += 1
            if end_parens_count == start_parens_count:
                result.append(term)
                end_parens_count, start_parens_count, term = 0, 0, ""
            else:
                term += "]"
        elif start_parens_count > end_parens_count:
            term += x
        else:
            result.extend(x.strip(" ").split(" "))


    return result


import multiprocessing as mp
from math import floor
from pathlib import Path
from typing import Collection

import dendropy
from rich.progress import track

from gtdblib import log


def bootstrap_merge_replicates(
        input_tree: Path,
        output_tree: Path,
        replicate_trees: Collection[Path],
        cpus: int = 1
):
    """Calculates non-parametric bootstraps for monophyletic groups and
    transposes them to the input tree.

    :param input_tree: The tree to merge the bootstraps to.
    :param output_tree: The path to write the tree to.
    :param replicate_trees: The trees to calculate the bootstraps from.
    :param cpus: The number of processes to use.
    """

    # Validate arguments
    if not isinstance(input_tree, Path):
        raise ValueError(f'Input tree is not a Path object.')
    if not isinstance(output_tree, Path):
        raise ValueError(f'Output tree is not a Path object.')
    if not isinstance(replicate_trees, Collection):
        raise ValueError(f'Replicate trees is not a Collection object.')
    if not all(isinstance(x, Path) for x in replicate_trees):
        raise ValueError(f'Not all replicate trees are Path objects.')
    if cpus is None or not isinstance(cpus, int) or cpus < 1:
        cpus = 1

    # Load the main tree that support values will be added to
    tree = dendropy.Tree.get_from_path(
        str(input_tree),
        schema='newick',
        rooting='force-unrooted',
        preserve_underscores=True
    )

    # Seed the support values
    for node in tree.internal_nodes():
        node.label = 0
        node.nontrivial_splits = 0

    # Store a consistent ordering of the internal nodes
    input_internal_nodes = tuple(tree.internal_nodes(), )

    # Generate the descendant taxa for those nodes
    input_desc_taxa = tuple([frozenset({y.taxon.label for y in x.leaf_nodes()}) for x in input_internal_nodes], )

    # Create a queue containing the input descendant taxa and
    queue = list()
    for rep_tree_path in replicate_trees:
        queue.append((rep_tree_path, input_desc_taxa))

    # Calculate the support values single-threaded
    if cpus == 1:
        results = [_calculate_support_worker(x) for x in track(queue, description='Processing...')]

    # Calculate support values with multiple threads
    else:
        with mp.Pool(processes=cpus) as pool:
            results = list(track(
                pool.imap_unordered(_calculate_support_worker, queue),
                description='Processing...',
                total=len(queue)
            ))

    # Append the split information to the nodes
    for result in results:
        for node, (n_support, n_non_trivial_split) in zip(input_internal_nodes, result):
            node.label += n_support
            node.nontrivial_splits += n_non_trivial_split

    # Calculate the %
    for node in tree.internal_nodes():
        if node.nontrivial_splits > 0:
            node.label = str(int(floor(node.label * 100.0 / node.nontrivial_splits)))
        else:
            node.label = 'NA'

    tree.write_to_path(str(output_tree), schema='newick', suppress_rooting=True, unquoted_underscores=True)
    log.info(f'Wrote tree to: {output_tree}')

    return


def _calculate_support_worker(job):
    """A multiprocessing worker to calculate the support for a split."""
    rep_tree_path, taxa_labels = job

    # Load the tree
    rep_tree = dendropy.Tree.get_from_path(
        str(rep_tree_path),
        schema='newick',
        rooting='force-unrooted',
        preserve_underscores=True
    )

    # Calculate the descendant taxa for the reference tree
    rep_tree_list = dendropy.TreeList([rep_tree])
    rep_tree_taxa_set = frozenset({x.taxon.label for x in rep_tree.leaf_node_iter()})

    # Iterate over the reference tree taxa internal nodes (order is consistent)
    results = list()
    for ref_taxa_labels in taxa_labels:

        # Only calculate the support for splits that are present in the reference tree
        taxa_labels = ref_taxa_labels.intersection(rep_tree_taxa_set)

        # Calculate the split
        split = rep_tree.taxon_namespace.taxa_bitmask(labels=taxa_labels)
        normalized_split = dendropy.Bipartition.normalize_bitmask(
            bitmask=split,
            fill_bitmask=rep_tree.taxon_namespace.all_taxa_bitmask(),
            lowest_relevant_bit=1)

        # Store the number of supported splits
        if len(taxa_labels) > 1:
            n_support = int(rep_tree_list.frequency_of_bipartition(split_bitmask=normalized_split))
            n_non_trivial_split = 1
        else:
            n_support = 0
            n_non_trivial_split = 0
        results.append((n_support, n_non_trivial_split))

    # Return the results (in the same order)
    return tuple(results, )



__title__ = 'gtdblib'
__description__ = 'An abstraction of objects, files, and third-party tools used by the GTDB. Not intended for public use.'
__url__ = 'https://github.com/Ecogenomics/gtdb-lib'
__version__ = '1.4.0'
__author__ = 'Aaron Mussig'
__author_email__ = 'aaronmussig@gmail.com'
__license__ = 'GPL-3.0'
__bug_url__ = 'https://github.com/Ecogenomics/gtdb-lib'
__doc_url__ = 'https://github.com/Ecogenomics/gtdb-lib'
__src_url__ = 'https://github.com/Ecogenomics/gtdb-lib'

import logging

from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET", format=FORMAT, datefmt="[%Y-%m-%d %H:%M:%S]", handlers=[RichHandler()]
)

log = logging.getLogger("rich")



from collections import defaultdict
from typing import Dict, FrozenSet

import dendropy

from gtdblib.tree.get_node_depth import get_tree_node_depth


def get_tree_node_to_desc_taxa(tree: dendropy.Tree) -> Dict[dendropy.Node, FrozenSet[str]]:
    """Get each node and the taxa that are descendants of this node.

    :param tree: The dendropy tree object to calculate the depths for.

    :returns: A dictionary of nodes and their descendant taxa.
    """

    # Calculate the nodes at each depth
    node_depth = get_tree_node_depth(tree)

    # Process the deepest nodes first
    out = defaultdict(set)
    for depth, nodes in sorted(node_depth.items(), key=lambda x: x[0], reverse=True):
        for node in nodes:

            # Seed the output dictionary
            if node.is_leaf():
                out[node].add(node.taxon.label)

            # Check bring up the taxa from the descendant nodes
            else:
                for child in node.child_nodes():
                    out[node].update(out[child])

    assert (out[tree.seed_node] == {x.label for x in tree.taxon_namespace})
    return {k: frozenset(v) for k, v in out.items()}



from pathlib import Path
from typing import List

import typer

from gtdblib.tree.bootstrap_merge import bootstrap_merge_replicates
from gtdblib.tree.polytomy import collapse_polytomy

app = typer.Typer()


@app.command()
def bootstrap_merge(
        tree: Path = typer.Argument(
            ...,
            help='Path to the Newick tree that bootstrap values will be added to.'
        ),
        output: Path = typer.Argument(
            ...,
            help='Path to the output tree.'
        ),
        replicates: List[Path] = typer.Argument(
            ...,
            help='Path(s) to the replicate tree(s).'
        ),
):
    """
    Calculate non-parametric bootstrap values using multiple trees.
    """
    bootstrap_merge_replicates(tree, output, replicates)


@app.command()
def polytomy(
        tree: Path = typer.Argument(
            ...,
            help='Path to the Newick tree.'
        ),
        output: Path = typer.Argument(
            ...,
            help='Path to the output tree.'
        ),
        support: float = typer.Argument(
            ...,
            min=0.0,
            max=100.0,
            help='Collapse nodes that have a support [red]less than[/red] this value.'
        ),
):
    """
    Collapse nodes with low bootstrap support.
    """
    collapse_polytomy(tree, output, support)



from pathlib import Path

import dendropy

from gtdblib import log
from gtdblib.exception import GtdbLibExit
from gtdblib.util.bio.newick import parse_label
from gtdblib.util.types import is_float


def _validate_support(support: float):
    """Validate that the support is a valid object."""
    if support is None or not is_float(support):
        raise GtdbLibExit(f'Support value "{support}" is not a float.')
    if support < 0.0 or support > 100.0:
        raise GtdbLibExit(f'Invalid support "{support}", must be between 0.0 and 100.0.')


def _validate_tree_path(path: Path):
    """Validate that the path is a valid tree path."""
    if path is None:
        raise GtdbLibExit(f'Path cannot be None.')
    if not isinstance(path, Path):
        raise GtdbLibExit(f'Path "{path}" is not a Path object.')


def _validate_tree_obj(tree: dendropy.Tree):
    if tree is None:
        raise GtdbLibExit(f'Tree cannot be None.')
    if not isinstance(tree, dendropy.Tree):
        raise GtdbLibExit(f'Tree "{tree}" is not a dendropy.Tree object.')
    if len(tree.taxon_namespace) == 0:
        raise GtdbLibExit(f'Tree "{tree}" has no taxa.')


def collapse_polytomy_tree(tree: dendropy.Tree, support: float):
    """Collapse nodes with low bootstrap support, works on the Tree object.

    :param tree: Tree to collapse.
    :param support: Collapse nodes that have a support less than this value.
    """
    log.info(f"Collapsing nodes with bootstrap support < {support}")

    # Validate arguments
    _validate_tree_obj(tree)
    _validate_support(support)

    n_internal_nodes = 0
    n_collapsed = 0
    for node in tree.internal_nodes():
        if node.label is not None:
            bs, _, _ = parse_label(node.label)
            if bs is not None and bs < support:
                for child_node in node.child_nodes():
                    child_node.edge_length += node.edge_length
                    node.remove_child(child_node)
                    node.parent_node.add_child(child_node)
                node.parent_node.remove_child(node)
                n_collapsed += 1
        n_internal_nodes += 1

    if n_internal_nodes > 0:
        log.info(f'Found {n_collapsed:,}/{n_internal_nodes:,} '
                 f'({n_collapsed / n_internal_nodes:.2%}) internal nodes '
                 f'to collapse.')
    else:
        raise GtdbLibExit(f'Did not find any nodes in the tree.')


def collapse_polytomy(path_in: Path, path_out: Path, support: float):
    """
    Collapse nodes with low bootstrap support.

    :param path_in: Tree to collapse.
    :param path_out: The path to write the tree to.
    :param support: Collapse nodes that have a support less than this value.
    """

    # Validate the arguments
    _validate_tree_path(path_in)
    _validate_tree_path(path_out)
    _validate_support(support)

    # Create the output directory if not exists
    path_out.parent.mkdir(parents=True, exist_ok=True)

    # Load the tree
    tree: dendropy.Tree = dendropy.Tree.get_from_path(
        str(path_in),
        schema='newick',
        rooting="force-unrooted",
        preserve_underscores=True
    )

    # Collapse nodes
    collapse_polytomy_tree(tree, support)

    # Write the output
    with open(path_out, 'w') as f:
        f.write(tree.as_string(schema='newick', suppress_rooting=True, unquoted_underscores=True))
    log.info(f'Wrote tree to: {path_out}')

    return



from gtdblib.taxon.rank import TaxonRank


class Taxon:
    __slots__ = ('rank', 'value')

    def __init__(self, rank: TaxonRank, value: str):
        self.rank: TaxonRank = rank
        self.value = value



from pathlib import Path
from typing import List, Literal

from gtdblib.file import File


class iTolDatasetTextFile(File):
    __slots__ = ('data',)

    def __init__(
            self,
            path: Path,
            label: str,
            color: str,
            margin: int = 0,
            show_internal: bool = False,
            label_rotation: float = 0,
            labels_below: bool = True,
            vertical_shift: int = 0,
            straight_labels: bool = False,
            align_to_tree: bool = False,
            size_factor: float = 1,
            external_label_shift: int = 0,
            show_labels: bool = False,
            label_size_factor: float = 1,
            label_shift: int = 0,
    ):
        """
        :param label: Used in the legend table.
        :param color: Used in the legend table.
        :param margin: left margin, used to increase/decrease the spacing to the next dataset. Can be negative, causing datasets to overlap. Used only for text labels which are displayed on the outside
        :param show_internal: applies to external text labels only; if set, text labels associated to internal nodes will be displayed even if these nodes are not collapsed. It could cause overlapping in the dataset display.
        :param label_rotation: Rotate all labels by the specified angle
        :param labels_below: By default, internal labels will be placed above the branches. If LABELS_BELOW is set to 1, labels will be below the branches
        :param vertical_shift: Shift internal labels vertically by this amount of pixels (positive or negative)
        :param straight_labels: If set to 1, tree rotation will not influence the individual label rotation
        :param align_to_tree: applies to external text labels only; If set to 1, labels will be displayed in arcs aligned to the tree (in circular mode) or vertically (in normal mode). All rotation parameters (global or individual) will be ignored.
        :param size_factor: font size factor; For external text labels, default font size will be slightly less than the available space between leaves, but you can set a multiplication factor here to increase/decrease it (values from 0 to 1 will decrease it, values above 1 will increase it)
        :param external_label_shift: add extra horizontal shift to the external labels. Useful in unrooted display mode to shift text labels further away from the node labels.
        :param show_labels: display or hide the dataset label above the external labels column
        :param label_size_factor: dataset label size factor
        :param label_shift: dataset label shift in pixels (positive or negative)
        """
        super().__init__(path)
        self.data: List[str] = [
            'DATASET_TEXT',
            'SEPARATOR TAB',
            f'DATASET_LABEL\t{label}',
            f'COLOR\t{color}',
            f'MARGIN\t{margin}',
            f'SHOW_INTERNAL\t{int(show_internal)}',
            f'LABEL_ROTATION\t{label_rotation}',
            f'LABELS_BELOW\t{int(labels_below)}',
            f'VERTICAL_SHIFT\t{vertical_shift}',
            f'STRAIGHT_LABELS\t{int(straight_labels)}',
            f'ALIGN_TO_TREE\t{int(align_to_tree)}',
            f'SIZE_FACTOR\t{size_factor}',
            f'EXTERNAL_LABEL_SHIFT\t{external_label_shift}',
            f'SHOW_LABELS\t{int(show_labels)}',
            f'LABEL_SIZE_FACTOR\t{label_size_factor}',
            f'LABEL_SHIFT\t{label_shift}',
            'DATA'
        ]

    def insert(
            self,
            node_id: str,
            label: str,
            position: float,
            color: str,
            style: Literal['normal', 'bold', 'italic', 'bold-italic'],
            size_factor: float,
            rotation: float
    ):
        """

        #position defines the position of the text label on the tree:
#  -1 = external label
#  a number between 0 and 1 = internal label positioned at the specified valu
e along the node branch (for example, position 0 is exactly at
the start of node branch, position 0.5 is in the middle, and position 1 is at the end)

'normal',''bold','italic' or 'bold-italic'
        """
        # #ID,label,position,color,style,size_factor,rotation

        if not (position == -1 or (0 <= position <= 1)):
            raise ValueError(f'Invalid position: {position}')
        self.data.append(f'{node_id}\t{label}\t{position}\t{color}\t{style}\t{size_factor}\t{rotation}')

    def write(self):
        with self.path.open('w') as f:
            f.write('\n'.join(self.data) + '\n')



from gtdblib.taxon.taxon import Taxon


class Taxonomy:
    __slots__ = ('d', 'p', 'c', 'o', 'f', 'g', 's')

    def __init__(self, d: Taxon, p: Taxon, c: Taxon, o: Taxon, f: Taxon, g: Taxon, s: Taxon):
        self.d = d
        self.p = p
        self.c = c
        self.o = o
        self.f = f
        self.g = g
        self.s = s



"""Utility functions for modifying color formats."""


def rgb_to_hex(r: int, g: int, b: int) -> str:
    """Convert RGB to hex color format.

    :param r: Red value.
    :param g: Green value.
    :param b: Blue value.

    :return: Hex color format.
    """
    return f'#{r:X}{g:X}{b:X}'

def hex_to_rgb(hex: str) -> tuple:
    """Convert hex to RGB color format.

    :param hex: Hex color format.

    :return: RGB color format.
    """
    return tuple(int(hex[i:i+2], 16) for i in (0, 2, 4))



"""Utility functions for copying and archiving files and directory trees."""
import errno
import logging
import os
import sys
import tempfile
import mmap


def get_num_lines(file_path):
    """ Calculate the number of lines in a file.

    :param file_path: path to file

    :return: number of lines in file
    """
    fp = open(file_path, "r+")
    buf = mmap.mmap(fp.fileno(), 0)
    lines = 0
    while buf.readline():
        lines += 1
    return lines



def symlink(target: str, link_name: str, overwrite: bool=False):
    '''Create a symbolic link named link_name pointing to target.

    If link_name exists then FileExistsError is raised, unless ``overwrite=True``.
    When trying to overwrite a directory, IsADirectoryError is raised.

    :param target: The target of the link.
    :param link_name: The name of the link.
    :param overwrite: If True, overwrite existing files and directories.

    :raises FileExistsError: If link_name exists and overwrite=False.

    :return: True

    '''

    if not overwrite:
        os.symlink(target, link_name)
        return

    # os.replace() may fail if files are on different filesystems
    link_dir = os.path.dirname(link_name)

    # Create link to target with temporary filename
    while True:
        temp_link_name = tempfile.mktemp(dir=link_dir)

        # os.* functions mimic as closely as possible system functions
        # The POSIX symlink() returns EEXIST if link_name already exists
        # https://pubs.opengroup.org/onlinepubs/9699919799/functions/symlink.html
        try:
            os.symlink(target, temp_link_name)
            break
        except FileExistsError:
            pass
    return True

def check_file_exists(input_file):
    """Check if file exists.This function is copied from `biolib <https://github.com/dparks1134/biolib>`_ on the 09/10/2020.

    :param input_file: path to file

    :return: True if path exists
    """

    if not os.path.exists(input_file) or not os.path.isfile(input_file):
        logger = logging.getLogger('timestamp')
        logger.error('Input file does not exists: ' + input_file + '\n')
        sys.exit()
    return True


def check_dir_exists(input_dir):
    """Check if directory exists.This function is copied from `biolib <https://github.com/dparks1134/biolib>`_ on the 09/10/2020.

    :param input_dir: path to directory

    :return: True if path exists
    """
    if not os.path.exists(input_dir) or not os.path.isdir(input_dir):
        logger = logging.getLogger('timestamp')
        logger.error('Input directory does not exists: ' + input_dir + '\n')
        sys.exit()
    return True


def make_sure_path_exists(path):
    """Create directory if it does not exist. This function is copied from `biolib <https://github.com/dparks1134/biolib>`_ on the 09/10/2020.

    :param path: path to directory

    :return: True if directory exists or was created

    """

    if not path:
        # lack of a path qualifier is acceptable as this
        # simply specifies the current directory
        return

    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            logger = logging.getLogger('timestamp')
            logger.error('Specified path could not be created: ' + path + '\n')
            sys.exit()
    return True



from typing import Tuple, Optional

from gtdblib.util.types import is_float


def parse_label(label: str) -> Tuple[Optional[float], Optional[str], Optional[str]]:
    """Parse a Newick label which may contain a support value, taxon, and/or auxiliary information.

    :param label: The label to parse.
    :returns: A tuple containing the support value, taxon, and auxiliary information.
    """

    support = None
    taxon = None
    auxiliary_info = None

    if label:
        label = label.strip()
        if '|' in label:
            label, auxiliary_info = label.split('|', 1)

        if ':' in label:
            support, taxon = label.split(':')
            support = float(support)
        else:
            if is_float(label):
                support = float(label)
            elif label != '':
                taxon = label

    return support, taxon, auxiliary_info


def create_label(support: Optional[float], taxon: Optional[str], auxiliary_info: Optional[str]) -> str:
    """Create label for Newick tree.

    :param support: The support value.
    :param taxon: The taxon.
    :param auxiliary_info: The auxiliary information.
    """

    label = ''
    if support is not None and taxon:
        label = str(support) + ':' + taxon
    elif support is not None:
        label = str(support)
    elif taxon:
        label = taxon

    if auxiliary_info:
        label += '|' + auxiliary_info

    return label



from pathlib import Path
from typing import List

from gtdblib.file import File


class iTolPopupFile(File):
    __slots__ = ('data',)

    def __init__(self, path: Path):
        super().__init__(path)
        self.data: List[str] = list()

    def insert(self, node_id: str, title: str, content: str):
        self.data.append(f'{node_id}\t{title}\t{content}')

    def write(self):
        with self.path.open('w') as f:
            f.write('\n'.join([
                                  'POPUP_INFO',
                                  'SEPARATOR TAB',
                                  'DATA'
                              ] + self.data) + '\n')



from pathlib import Path
from typing import List

from gtdblib.file import File


class iTolCollapseFile(File):
    __slots__ = ('data',)

    def __init__(self, path: Path):
        super().__init__(path)
        self.data: List[str] = list()

    def insert(self, node_id: str):
        self.data.append(node_id)

    def write(self):

        with self.path.open('w') as f:
            f.write('\n'.join([
                'COLLAPSE',
                'DATA'
            ] + self.data) + '\n')



from pathlib import Path
from typing import List, Literal

from gtdblib.file import File


class iTolDatasetColorStripFile(File):
    __slots__ = ('data',)

    def __init__(
            self,
            path: Path,
            label: str,
            color: str,
            color_branches: bool = False,
            strip_width: int = 25,
            margin: int = 0,
            border_width: int = 0,
            border_color: str = '#0000ff',
            complete_border: bool = False,
            show_internal: bool = False,
            show_strip_labels: bool = True,
            strip_label_position: Literal['top', 'center', 'bottom'] = 'center',
            strip_label_size_factor: float = 1,
            strip_label_rotation: float = 0,
            strip_label_shift: int = 0,
            strip_label_color: str = '#000000',
            strip_label_outline: float = 0.5,
            show_labels: bool = False,
            size_factor: float = 1,
            label_rotation: float = 0,
            label_shift: float = 0,
    ):
        """
        """
        super().__init__(path)
        self.data: List[str] = [
            'DATASET_COLORSTRIP',
            'SEPARATOR TAB',
            f'DATASET_LABEL\t{label}',
            f'COLOR\t{color}',
            f'COLOR_BRANCHES\t{int(color_branches)}',
            f'STRIP_WIDTH\t{strip_width}',
            f'MARGIN\t{margin}',
            f'BORDER_WIDTH\t{border_width}',
            f'BORDER_COLOR\t{border_color}',
            f'COMPLETE_BORDER\t{int(complete_border)}',
            f'SHOW_INTERNAL\t{int(show_internal)}',
            f'SHOW_STRIP_LABELS\t{int(show_strip_labels)}',
            f'STRIP_LABEL_POSITION\t{strip_label_position}',
            f'STRIP_LABEL_SIZE_FACTOR\t{strip_label_size_factor}',
            f'STRIP_LABEL_ROTATION\t{strip_label_rotation}',
            f'STRIP_LABEL_SHIFT\t{strip_label_shift}',
            f'STRIP_LABEL_COLOR\t{strip_label_color}',
            f'STRIP_LABEL_OUTLINE\t{strip_label_outline}',
            f'SHOW_LABELS\t{int(show_labels)}',
            f'SIZE_FACTOR\t{size_factor}',
            f'LABEL_ROTATION\t{label_rotation}',
            f'LABEL_SHIFT\t{label_shift}',
            'DATA'
        ]

    def insert(self, node_id: str, color: str, label: str, ):
        self.data.append(f'{node_id}\t{color}\t{label}')

    def write(self):
        with self.path.open('w') as f:
            f.write('\n'.join(self.data) + '\n')



from collections import deque, defaultdict
from typing import Dict, Set

import dendropy


def get_tree_node_depth(tree: dendropy.Tree) -> Dict[int, Set[dendropy.Node]]:
    """Calculate the depth of all nodes in the tree.

    :param tree: The dendropy tree object to calculate the depths for.

    :returns: A dictionary of depths to nodes at that depth.
    """
    # Validate arguments
    if not isinstance(tree, dendropy.Tree):
        raise ValueError(f'Tree is not a dendropy.Tree object.')

    # Create a queue starting from the seed node (depth=0)
    out = defaultdict(set)
    queue = deque([(tree.seed_node, 0)])

    # Progressively add children to the queue
    while len(queue) > 0:
        cur_node, cur_depth = queue.popleft()
        out[cur_depth].add(cur_node)

        for child in cur_node.child_nodes():
            queue.append((child, cur_depth + 1))

    # Make this a dictionary (not defaultdict)
    return dict(out)



def is_float(s) -> bool:
    """Check if a string can be converted to a float.

    :param s: Any object to convert to a float.
    """
    try:
        float(s)
    except ValueError:
        return False
    return True



from pathlib import Path

from gtdblib.exception import GtdbLibException


class File:
    """The base class that all files inherit from."""

    __slots__ = ('path',)

    def __init__(self, path: Path):
        # Validate arguments
        if not isinstance(path, Path):
            raise GtdbLibException("path must be a pathlib.Path object")
        if path.is_dir():
            raise GtdbLibException(f"path must be a file, not a directory: {path}")

        self.path: Path = path



import tempfile
import unittest
from pathlib import Path

import dendropy

from gtdblib.tree.bootstrap_merge import bootstrap_merge_replicates


def _create_test_tree(a: str, b: str, c: str, d: str, path: Path):
    """
    #
    #         (seed)
    #      /          \
    #    ch1          ch2
    #   /  \          /  \
    #  ch3  ch4     ch5  ch6
    """
    taxon_namespace = dendropy.TaxonNamespace(["A", "B", "C", "D", ])
    tree = dendropy.Tree(taxon_namespace=taxon_namespace)

    # Left branch
    ch1 = tree.seed_node.new_child(edge_length=1)
    ch3 = ch1.new_child(edge_length=1)
    ch4 = ch1.new_child(edge_length=1)

    # Right branch
    ch2 = tree.seed_node.new_child(edge_length=1)
    ch5 = ch2.new_child(edge_length=2)
    ch6 = ch2.new_child(edge_length=3)

    # Assign taxa
    ch3.taxon = taxon_namespace.get_taxon(a)
    ch4.taxon = taxon_namespace.get_taxon(b)
    ch5.taxon = taxon_namespace.get_taxon(c)
    ch6.taxon = taxon_namespace.get_taxon(d)

    tree.write_to_path(str(path),
                       schema='newick',
                       suppress_rooting=True,
                       unquoted_underscores=True)

    return tree


def _create_test_tree_2(a: str, b: str, c: str, d: str, path: Path):
    """
    #
    #         (seed)
    #      /          \
    #    ch1          ch2
    #             /         \
    #           ch3         ch4
                          /      \
                        ch5      ch6
    """
    taxon_namespace = dendropy.TaxonNamespace(["A", "B", "C", "D", ])
    tree = dendropy.Tree(taxon_namespace=taxon_namespace)

    # Left branch
    ch1 = tree.seed_node.new_child(edge_length=1)

    # Right branch
    ch2 = tree.seed_node.new_child(edge_length=1)

    ch3 = ch2.new_child(edge_length=1)
    ch4 = ch2.new_child(edge_length=1)

    ch5 = ch4.new_child(edge_length=2)
    ch6 = ch4.new_child(edge_length=3)

    # Assign taxa
    ch1.taxon = taxon_namespace.get_taxon(a)
    ch3.taxon = taxon_namespace.get_taxon(b)
    ch5.taxon = taxon_namespace.get_taxon(c)
    ch6.taxon = taxon_namespace.get_taxon(d)

    tree.write_to_path(str(path),
                       schema='newick',
                       suppress_rooting=True,
                       unquoted_underscores=True)

    return tree


class TestTreeBootstrapMerge(unittest.TestCase):

    # def test_bootstrap_merge_replicates(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         tmp_dir = Path(tmp_dir)
    #
    #         path_ref = tmp_dir / 'ref.tree'
    #         path_1 = tmp_dir / '1.tree'
    #         path_2 = tmp_dir / '2.tree'
    #         path_3 = tmp_dir / '3.tree'
    #
    #         _create_test_tree('A', 'B', 'C', 'D', path_ref)
    #
    #         _create_test_tree('A', 'D', 'B', 'C', path_1)
    #         _create_test_tree('B', 'A', 'C', 'D', path_2)
    #         _create_test_tree_2('A', 'B', 'C', 'D', path_3)
    #
    #         path_out = tmp_dir / 'out.tree'
    #         bootstrap_merge_replicates(path_ref, path_out, [path_1, path_2, path_3])
    #
    #         tree = dendropy.Tree.get_from_path(str(path_out), schema='newick')
    #
    #         seed_bs = float(tree.seed_node.label)
    #
    #         self.assertEqual(seed_bs, 100.0)

   def test_bootstrap_merge_replicates_local(self):
        path_out = Path('/tmp/ar/merged.tree')
        path_ref = Path('/tmp/ar/ar53_r207.tree')

        import os
        rep_paths = list()
        for file in os.listdir('/tmp/ar/bootstrap_trees'):
            if file.endswith('.treefile'):
                rep_paths.append(Path('/tmp/ar/bootstrap_trees') / file)

        rep_paths = rep_paths[0:2]
        bootstrap_merge_replicates(path_ref, path_out, rep_paths, cpus=5)
        return






import tempfile
import unittest
from collections import defaultdict
from pathlib import Path

import dendropy

from gtdblib.tree.polytomy import collapse_polytomy


def _create_test_tree():
    """
    #
    #          (seed)
    #       /        \
    #      /          \
    #    ch1 [80]     ch2 [50]
    #   /  \          /  \
    #  ch3 ch4       ch5 ch6
    """
    taxon_namespace = dendropy.TaxonNamespace(["A", "B", "C", "D", ])
    tree = dendropy.Tree(taxon_namespace=taxon_namespace)

    # Left branch
    ch1 = tree.seed_node.new_child(edge_length=1)
    ch1.label = '80'

    ch3 = ch1.new_child(edge_length=1.5)
    ch4 = ch1.new_child(edge_length=2.3)

    # Right branch
    ch2 = tree.seed_node.new_child(edge_length=1.26)
    ch2.label = '50'

    ch5 = ch2.new_child(edge_length=3.1)
    ch6 = ch2.new_child(edge_length=2.9)

    # Assign taxa
    ch3.taxon = taxon_namespace.get_taxon("A")
    ch4.taxon = taxon_namespace.get_taxon("B")
    ch5.taxon = taxon_namespace.get_taxon("C")
    ch6.taxon = taxon_namespace.get_taxon("D")

    return tree


class TestTreePolytomy(unittest.TestCase):

    def test_collapse_polytomy(self):

        tree = _create_test_tree()

        with tempfile.TemporaryDirectory() as tmp_dir:
            path_tmp = Path(tmp_dir) / 'test.tree'
            with open(path_tmp, 'w') as f:
                f.write(tree.as_string(schema='newick', suppress_rooting=True, unquoted_underscores=True))

            path_out = Path(tmp_dir) / 'out.tree'
            collapse_polytomy(path_tmp, path_out, 70)

            tree_collapsed = dendropy.Tree.get_from_path(str(path_out), schema='newick', rooting='force-rooted',
                                                         preserve_underscores=True)

        self.assertEqual(str(tree_collapsed), '((A:1.5,B:2.3)80:1.0,C:4.36,D:4.16)')

        pdm_a = tree.phylogenetic_distance_matrix()
        pdm_b = tree_collapsed.phylogenetic_distance_matrix()

        pdm_a_values = defaultdict(dict)
        for t1 in tree.taxon_namespace:
            for t2 in tree.taxon_namespace:
                pdm_a_values[t1.label][t2.label] = (pdm_a.patristic_distance(t1, t2))

        pdm_b_values = defaultdict(dict)
        for t1 in tree_collapsed.taxon_namespace:
            for t2 in tree_collapsed.taxon_namespace:
                pdm_b_values[t1.label][t2.label] = (pdm_b.patristic_distance(t1, t2))

        self.assertAlmostEqual(pdm_a_values['A']['B'], pdm_b_values['A']['B'])
        self.assertAlmostEqual(pdm_b_values['A']['C'], 6.86)
        self.assertAlmostEqual(pdm_b_values['A']['D'], 6.66)
        self.assertAlmostEqual(pdm_b_values['B']['C'], 7.66)
        self.assertAlmostEqual(pdm_b_values['B']['D'], 7.46)



from typing import Dict, Tuple

import dendropy


def get_tree_unique_mrca_for_internal_nodes(tree: dendropy.Tree) -> Dict[dendropy.Node, Tuple[str, str]]:
    """Calculate a unique MRCA for each internal node, i.e. using these taxa
    will always yield the same internal node.

    :param tree: The dendropy tree object to calculate the depths for.

    :returns: A dictionary of nodes and a tuple of each taxon that identifies the node.
    """

    out = dict()
    for node in tree.internal_nodes():
        child_nodes = node.child_nodes()
        if len(child_nodes) < 2:
            raise ValueError("Node must have at least 2 children")
        left_child = child_nodes[0]
        right_child = child_nodes[1]

        left_taxon = left_child.leaf_nodes()[0].taxon.label
        right_taxon = right_child.leaf_nodes()[0].taxon.label

        out[node] = (left_taxon, right_taxon)

    return out



from enum import Enum


class TaxonRank(Enum):
    """A taxonomic rank (e.g. Domain, Phylum, etc...)."""

    DOMAIN = 'd__'
    PHYLUM = 'p__'
    CLASS = 'c__'
    ORDER = 'o__'
    FAMILY = 'f__'
    GENUS = 'g__'
    SPECIES = 's__'



import typer

from gtdblib import __version__, log
from gtdblib.cli.tree import app as tree_app
from gtdblib.exception import GtdbLibExit

app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]}, rich_markup_mode="rich")


# The section below adds global commands to the app
def version_callback(value: bool):
    if value:
        typer.echo(f"gtdblib: {__version__}")
        raise typer.Exit()


@app.callback()
def common(
        ctx: typer.Context,
        version: bool = typer.Option(None, "--version", "-v", callback=version_callback),
):
    pass


# Add subparsers
app.add_typer(tree_app, name="tree", help='Commands for working with Newick trees.')


def main():
    """
    The main method that wraps all logic for the CLI.
    """
    try:
        app()
    except GtdbLibExit as e:
        log.error(e)
        typer.Exit(1)


if __name__ == "__main__":
    main()



import os
import re

from setuptools import setup, find_packages


def read_meta():
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'gtdblib/__init__.py')
    with open(path) as fh:
        hits = re.findall(r'__(\w+)__ ?= ?["\'](.+)["\']\n', fh.read())
    return {k: v for k, v in hits}


def readme():
    with open('README.md') as f:
        return f.read()


meta = read_meta()
setup(name=meta['title'],
      version=meta['version'],
      description=meta['description'],
      long_description=readme(),
      long_description_content_type='text/markdown',
      author=meta['author'],
      author_email=meta['author_email'],
      url=meta['url'],
      license=meta['license'],
      project_urls={
          'Bug Tracker': meta['bug_url'],
          'Documentation': meta['doc_url'],
          'Source Code': meta['src_url'],
      },
      entry_points={
          'console_scripts': [
              'gtdblib = gtdblib.__main__:main'
          ]
      },
      classifiers=[
          'Development Status :: 1 - Planning',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Topic :: Software Development',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
      packages=find_packages(),
      include_package_data=True,
      install_requires=['numpy>=1.9.0', 'tqdm>=4.35.0', 'typer[all]', 'dendropy'],
      python_requires='>=3.6',
      )



from pathlib import Path
from typing import List, Optional

from gtdblib.file import File


class iTolTreeColorsFile(File):
    __slots__ = ('data',)

    def __init__(self, path: Path):
        super().__init__(path)
        self.data: List[str] = list()

    def insert_range(
            self, node_id: str, color: str, label: str):
        """defines a colored range (colored background for labels/clade)"""
        self.data.append('\t'.join([node_id, 'range', color, label]))

    def insert_clade(
            self,
            node_id: str,
            color: str,
            branch_style: str,  # Literal['normal', 'dashed']
            scale_factor: float):
        """defines color/style for all branches in a clade"""
        self.data.append('\t'.join([node_id, 'clade', color, branch_style, str(scale_factor)]))

    def insert_branch(
            self,
            node_id: str,
            color: str,
            branch_style: str,  # Literal['normal', 'dashed']
            scale_factor: float):
        """defines color/style for a single branch"""
        self.data.append('\t'.join([node_id, 'branch', color, branch_style, str(scale_factor)]))

    def insert_label(
            self,
            node_id: str,
            color: str,
            font_style: Optional[str] = None,  # Literal['normal', 'bold', 'italic', 'bold-italic']
            scale_factor: Optional[float] = None,
    ):
        """defines font color/style for the leaf label"""
        row = [node_id, 'label', color]
        if font_style is not None:
            row.append(font_style)
        if scale_factor is not None:
            row.append(str(scale_factor))
        self.data.append('\t'.join(row))

    def insert_label_background(self, node_id: str, color: str):
        """defines the leaf label background color"""
        self.data.append('\t'.join([node_id, 'label_background', color]))

    def write(self):
        with self.path.open('w') as f:
            f.write('\n'.join([
                                  'TREE_COLORS',
                                  'SEPARATOR TAB',
                                  'DATA'
                              ] + self.data) + '\n')



class GtdbLibException(Exception):
    """Base exception for all gtdb-lib exceptions thrown in this project."""

    def __init__(self, message: str = ''):
        Exception.__init__(self, message)

class GtdbLibExit(GtdbLibException):
    """Raised when gtdblib has been asked to quietly exit."""

    def __init__(self, message: str = ''):
        GtdbLibException.__init__(self, message)




###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################


import os
import sys
import gzip
import traceback

from .exceptions import BioError


class InputFileError(BioError):
    pass


protein_bases = {'a', 'r', 'n', 'd', 'c', 'q', 'e', 'g', 'h', 'i', 'l', 'k', 'm', 'f', 'p', 's', 't', 'w', 'y', 'v'}
nucleotide_bases = {'a', 'c', 'g', 't'}
insertion_bases = {'-', '.'}


def is_nucleotide(seq_file, req_perc=0.95, max_seqs_to_read=10):
    '''Check if a file contains sequences in nucleotide space.

    The check is performed by looking for the characters in
    {a,c,g,t,n,.,-} and confirming that these comprise the
    majority of a sequences. A set number of sequences are
    read and the file assumed to be not be in nucleotide space
    if none of these sequences are comprised primarily of the
    defined nucleotide set.

    :param seq_file: str, Name of fasta/q file to read.
    :param req_perc: float, Percentage of nucleotide bases before declaring the sequences as being in nucleotide space.
    :param max_seqs_to_read: int, Maximum sequences to read before declaring sequence file to not be in nucleotide space.

    :return: boolean, True is sequences are in nucleotide space.

    '''

    seq_count = 0
    for _seq_id, seq in read_seq(seq_file):
        seq = seq.lower()

        nt_bases = 0
        for c in (nucleotide_bases | {'n'} | insertion_bases):
            nt_bases += seq.count(c)

        if nt_bases / len(seq) >= req_perc:
            return True

        seq_count += 1
        if seq_count == max_seqs_to_read:
            break

    return False


def is_protein(seq_file, req_perc=0.95, max_seqs_to_read=10):
    """Check if a file contains sequences in protein space.

    The check is performed by looking for the 20 amino acids,
    along with X, and the insertion characters '-' and '.', in
    order to confirm that these comprise the majority of a
    sequences. A set number of sequences are read and the file
    assumed to be not be in nucleotide space if none of these
    sequences are comprised primarily of the defined nucleotide set.

    :param seq_file: str, Name of fasta/q file to read.
    :param req_perc: float, Percentage of amino acid bases before declaring the sequences as being in amino acid space.
    :param max_seqs_to_read: int, Maximum sequences to read before declaring sequence file to not be in amino acid space.

    :return: boolean, True is sequences are in amino acid space.

    """

    seq_count = 0
    for _seq_id, seq in read_seq(seq_file):
        seq = seq.lower()

        prot_bases = 0
        for c in (protein_bases | {'x'} | insertion_bases):
            prot_bases += seq.count(c)

        if prot_bases / len(seq) >= req_perc:
            return True

        seq_count += 1
        if seq_count == max_seqs_to_read:
            break

    return False


def read(seq_file):
    """Read sequences from fasta/q file.

    :param seq_file: str, Name of fasta/q file to read.

    :return: dict[seq_id] -> seq, Sequences indexed by sequence id.

    """
    if seq_file.endswith(('.fq.gz', '.fastq.gz', '.fq', '.fq.gz')):
        return read_fastq(seq_file)
    else:
        return read_fasta(seq_file)


def read_fasta(fasta_file, keep_annotation=False):
    """Read sequences from fasta file.

    :param fasta_file: str, Name of fasta file to read.
    :param keep_annotation: boolean, Determine is sequence id should contain annotation.

    :return: dict[seq_id] -> seq, Sequences indexed by sequence id.
    """

    if not os.path.exists(fasta_file):
        raise InputFileError('Input file %s does not exist.' % fasta_file)

    if os.stat(fasta_file).st_size == 0:
        return {}

    try:
        open_file = open
        if fasta_file.endswith('.gz'):
            open_file = gzip.open

        seqs = {}
        for line in open_file(fasta_file, 'rt'):
            # skip blank lines
            if not line.strip():
                continue

            if line[0] == '>':
                if keep_annotation:
                    seq_id = line[1:-1]
                else:
                    seq_id = line[1:].split(None, 1)[0]

                seqs[seq_id] = []
            else:
                seqs[seq_id].append(line.strip())

        for seq_id, seq in seqs.items():
            seqs[seq_id] = ''.join(seq).replace(' ', '')
    except:
        print(traceback.format_exc())
        print()
        print("[Error] Failed to process sequence file: " + fasta_file)
        sys.exit(1)

    return seqs


def read_fastq(fastq_file):
    """Read sequences from fastq file.

    :param fastq_file: str, Name of fastq file to read.

    :return: dict[seq_id] -> seq, Sequences indexed by sequence id.
    """

    if not os.path.exists(fastq_file):
        raise InputFileError('Input file %s does not exist.' % fastq_file)

    if os.stat(fastq_file).st_size == 0:
        return {}

    try:
        open_file = open
        if fastq_file.endswith('.gz'):
            open_file = gzip.open

        seqs = {}
        line_num = 0
        for line in open_file(fastq_file, 'rt'):
            line_num += 1

            if line_num == 1:
                seq_id = line[1:].split(None, 1)[0]
            elif line_num == 2:
                seqs[seq_id].seq = line.strip()
            elif line_num == 4:
                line_num = 0
    except:
        print(traceback.format_exc())
        print()
        print("[Error] Failed to process sequence file: " + fastq_file)
        sys.exit(1)

    return seqs


def read_seq(seq_file, keep_annotation=False):
    """Generator function to read sequences from fasta/q file.

    This function is intended to be used as a generator
    in order to avoid having to have large sequence files
    in memory. Input file may be gzipped and in either
    fasta or fastq format. It is slightly more efficient
    to directly call read_fasta_seq() or read_fastq_seq()
    if the type of input file in known.

    :Example:

        >>> seq_io = SeqIO()
        >>> for seq_id, seq in seq_io.read_seq(fasta_file):
        >>>     print seq_id
        >>>>    print seq

    :param seq_file:str, Name of fasta/q file to read.
    :param keep_annotation:boolean, Determine if annotation string should be returned.

    :return: Iterator[seq_id, seq, [annotation]], Unique id of the sequence followed by the sequence itself,
        and the annotation if keep_annotation is True.

    """

    if seq_file.endswith(('.fq.gz', '.fastq.gz', '.fq', '.fq.gz')):
        for rtn in read_fastq_seq(seq_file):
            yield rtn
    else:
        for rtn in read_fasta_seq(seq_file, keep_annotation):
            yield rtn


def read_fasta_seq(fasta_file, keep_annotation=False):
    """Generator function to read sequences from fasta file.

    This function is intended to be used as a generator
    in order to avoid having to have large sequence files
    in memory. Input file may be gzipped.

    :Example:
        >>> seq_io = SeqIO()
        >>> for seq_id, seq in seq_io.read_fasta_seq(fasta_file):
        >>>     print seq_id
        >>>     print seq

    :param fasta_file: str, Name of fasta file to read.
    :param keep_annotation: boolean, Determine if annotation string should be returned.

    :return: Iterator[seq_id, seq, [annotation]], Unique id of the sequence followed by the sequence itself,
        and the annotation if keep_annotation is True.

    """

    if not os.path.exists(fasta_file):
        raise InputFileError('Input file %s does not exist.' % fasta_file)

    if os.stat(fasta_file).st_size == 0:
        pass

    try:
        open_file = open
        if fasta_file.endswith('.gz'):
            open_file = gzip.open

        seq_id = None
        annotation = None
        seq = None
        for line in open_file(fasta_file, 'rt'):
            # skip blank lines
            if not line.strip():
                continue

            if line[0] == '>':
                if seq_id != None:
                    if keep_annotation:
                        yield seq_id, ''.join(seq).replace(' ', ''), annotation
                    else:
                        yield seq_id, ''.join(seq).replace(' ', '')

                line_split = line[1:-1].split(None, 1)
                if len(line_split) == 2:
                    seq_id, annotation = line_split
                else:
                    seq_id = line_split[0]
                    annotation = ''
                seq = []
            else:
                seq.append(line.strip())

        # report last sequence
        if keep_annotation:
            yield seq_id, ''.join(seq).replace(' ', ''), annotation
        else:
            yield seq_id, ''.join(seq).replace(' ', '')
    except GeneratorExit:
        pass
    except:
        print(traceback.format_exc())
        print()
        print("[Error] Failed to process sequence file: " + fasta_file)
        sys.exit(1)


def read_fastq_seq(fastq_file):
    """Generator function to read sequences from fastq file.

    This function is intended to be used as a generator
    in order to avoid having to have large sequence files
    in memory. Input file may be gzipped.

    :Example:
        >>> seq_io = SeqIO()
        >>> for seq_id, seq in seq_io.read_fastq_seq(fastq_file):
        >>>     print seq_id
        >>>     print seq

    :param fastq_file: str, Name of fastq file to read.

    :return: Iterator[seq_id, seq], Unique id of the sequence followed by the sequence itself.

    """

    if not os.path.exists(fastq_file):
        raise InputFileError('Input file %s does not exist.' % fastq_file)

    if os.stat(fastq_file).st_size == 0:
        pass

    try:
        open_file = open
        if fastq_file.endswith('.gz'):
            open_file = gzip.open

        line_num = 0
        for line in open_file(fastq_file, 'rt'):
            line_num += 1

            if line_num == 1:
                seq_id = line[1:].split(None, 1)[0]
            elif line_num == 2:
                yield seq_id, line.strip()
            elif line_num == 4:
                line_num = 0
    except GeneratorExit:
        pass
    except:
        print(traceback.format_exc())
        print()
        print("[Error] Failed to process sequence file: " + fastq_file)
        sys.exit(1)


def extract_seqs(fasta_file, seqs_to_extract):
    """Extract specific sequences from fasta file.

    :param fasta_file: str, Fasta file containing sequences.
    :param seqs_to_extract: set, Ids of sequences to extract.

    :return: dict[seq_id] -> seq, Dictionary of sequences indexed by sequence id.

    """

    if not os.path.exists(fasta_file):
        raise InputFileError('Input file %s does not exist.' % fasta_file)

    if os.stat(fasta_file).st_size == 0:
        return {}

    seqs = {}
    for line in open(fasta_file):
        if line[0] == '>':
            seq_id = line[1:].partition(' ')[0]

            seq_of_interest = False
            if seq_id in seqs_to_extract:
                seqs[seq_id] = []
                seq_of_interest = True
        elif seq_of_interest:
            seqs[seq_id].append(line.strip())

    for seq_id, seq in seqs.items():
        seqs[seq_id] = ''.join(seq).replace(' ', '')

    return seqs


def seq_lengths(fasta_file):
    """Calculate length of each sequence.

    :param fasta_file: str, Fasta file containing sequences.

    :return: d[seq_id] -> length, Length of each sequence.

    """

    if not os.path.exists(fasta_file):
        raise InputFileError('Input file %s does not exist.' % fasta_file)

    if os.stat(fasta_file).st_size == 0:
        return {}

    lens = {}
    for seq_id, seq in read_fasta_seq(fasta_file):
        if seq[-1] == '*':
            lens[seq_id] = len(seq) - 1
        else:
            lens[seq_id] = len(seq)

    return lens


def write_fasta(seqs, output_file, wrap=80):
    """Write sequences to fasta file.

    If the output file has the extension 'gz',
    it will be compressed using gzip.

    :param seqs: dict[seq_id] -> seq, Sequences indexed by sequence id.
    :param output_file: str, Name of fasta file to produce.
    :param wrap: int, Number of characters per line.

    """

    if output_file.endswith('.gz'):
        fout = gzip.open(output_file, 'wb')
    else:
        fout = open(output_file, 'w')

    for gid, gseq in seqs.items():
        fout.write('>{}\n'.format(gid))
        for i in range(0, len(gseq), wrap):
            fout.write('{}\n'.format(gseq[i:i + wrap]))
    fout.close()
